#{----------------------------------------------------------------------------
# Copyright 2009, Jakub Kupikowski
#
# License: 
# Free for self use
# Permission to sell or redistribute or repackage this software 
# should be specifically requested from the author.
# above copyright notice should appear in all copies.
#
# THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#
# ----------------------------------------------------------------------------
#
# Author: Jakub Kupikowski
#
# Name: simple_rays.rb
#
# Description: Textures model according to chosen raytest method.
#		
# Usage   : Selection > Right Mouse Click > Raytest
#		 
# Install : Install rbz archive using "Install Extension..." utility, from:
#            Window (menu) > Preferences (menuitem) > Extensions (panel)
#
# ----------------------------------------------------------------------------
# VERSIONS:
#
#   0.1.0 : 2009-08-08 : Jakub Kupikowski (qpik)
#         + first release
#
#   0.2.0 : 2009-08-10 : Jakub Kupikowski (qpik)
#         ~ Modified ray casting algorithm - more even ray casting,
#             resulting also in approx. 2 times more rays cast.
#
#   0.3.0 : 2009-08-10 : Jakub Kupikowski (qpik)
#         ~ Modified lightmap algorithm - shading time reduced approx.
#             2 times by reducing number of occlusion calls.
#         + Added Bitmap.expand method to do the same task.
#         - Deleted some unused code.
#
#   0.4.0 : 2009-08-11 : Jakub Kupikowski (qpik)
#         + Added Bitmap.blur - texture smoothing.
#
#   0.5.0 : 2009-08-11 : Jakub Kupikowski (qpik)
#         + Added progress info.
#         - Removed unused parts of code.
#
#   0.6.0 : 2009-08-12 : Jakub Kupikowski (qpik)
#         - Removed Save panel.
#         ~ When model is not saved, "temp_occlusion" folder is created
#             in the Plugins folder
#         + Added Mac compatibility.
#
#   0.7.0 : 2009-11-25 : Jakub Kupikowski (qpik)
#         + Added Sun Exposure Time mode.
#         ~ Changed the name of the plugin.
#         ~ Rewrote many parts of code, making it a bit faster.
#
#   1.0.0 : 2012-12-31 : Daniel A. Rathbun
#         + Namespace wrapped within Qpik::SimpleRays
#         ~ Re-written to avoid modifying Trimble OEM API classes.
#         ~ Re-written to avoid modifying any Ruby base classes.
#         ~ Untitled models now use '~/_temp_simple_rays' as temp dir
#           (formerly used a temp dir in the application dir hierarchy.)
#         + Temporary image files and dirs are cleaned up (if not @@debug)
#
#   (See the file "v1.0.0_overhal.txt" for exhaustive listing of changes.)
#
#   1.0.1 : 2013-01-05 : Daniel A. Rathbun
#         ~ Correct @@lang assignment in language file rescue block.
#
#}----------------------------------------------------------------------------

module Qpik

  module SimpleRays

    require('sketchup.rb')

    # MODULE CONSTANTS
    BASEPATH = File.dirname(__FILE__) unless defined?(BASEPATH)
    VERSION  = '1.0.1' unless defined?(VERSION)

    #{# MODULE VARIABLES
    #
      @@debug = false unless defined?(@@debug)
      @@debug_state = MF_UNCHECKED unless defined?(@@debug_state)
      
      @@command = {} unless defined?(@@command)

      @@select_proc = Proc.new {
        Sketchup.active_model.selection.empty? ? MF_GRAYED : MF_ENABLED
      }

      @@prevDir = Dir.getwd unless defined?(@@prevDir)

      #{ String hashes:
      @@lang = Sketchup.get_locale()[0,2] unless defined?(@@lang)
      unless @@lang == 'en'
        # Load localized string hashes from file:
        begin
          load(File.join(BASEPATH,"simple_rays_"<<@@lang<<".rb"))
        rescue LoadError => e
          unless $VERBOSE.nil?
            puts()
            puts('NOTICE: Could not find a "simple_rays_'<<@@lang<<'.rb" file to load.')
            puts('Using English text for SimpleRays plugin UI.')
            puts()
          end
          @@lang = 'en' # just use English
        rescue => e
          @@lang = 'en' # just use English
        end
      end # load attempt
      #
      if @@lang == 'en'
        @@menutext = Hash[
          :submenu, "SimpleRays",
          #
          :simple_rays, "SimpleRays Options...",
          #
          :debug_mode, "Debug Mode"
        ]
        @@tooltips = Hash[
          :simple_rays, "Texture selected faces using simple rays"
        ]
        @@opt_text = Hash[
          :label_meth, "Method:",
          :label_prec, "Precision:",
          :label_rays, "Rays:",
          :meth_0, "Sun Exposure Time",
          :meth_1, "Ambient Occlusion"
        ]
      end #}

      @@labels = [
        @@opt_text[:label_meth],
        @@opt_text[:label_prec],
        @@opt_text[:label_rays]
      ]

      @@raymeths = "%s|%s" % [@@opt_text[:meth_0],@@opt_text[:meth_1]]

      @@def_opts = [@@opt_text[:meth_0],"10","10"] unless defined?(@@def_opts)

    #
    #}#


    class Bitmap

      attr_reader(:data, :width, :height, :size)
      attr_accessor(:path) ### v1.0.0

      #{ initialize()
      #
      def initialize(*params)
        if params.length == 2
          img_width = params[0]
          img_height = params[1]
          @data = []
          @width = img_width
          @height = img_height
          for j in 0..(@height-1)
            row = []
            for i in 0..(@width-1)
              row[i] = [0, 0, 0]
            end
            @data[j] = row
          end
          padding = @width.divmod(4)[1]
          @size = 54 + (@width * 3 + padding) * @height
        elsif params.length == 1
          array = params[0]
          @data = array
          @width = array[0].length
          @height = array.length
          padding = @width.divmod(4)[1]
          @size = 54 + (@width * 3 + padding) * @height
        end
      end #} initialize()

      #{ write()
      #
      def write(file_name)
        header = 0.chr*54
        header[0,2] = 'BM'
        header[2,4] = lbo_dword(@size) #? [@size].pack("V")
        header[10] = 54.chr
        header[14] = 40.chr
        header[18,4] = lbo_dword(@width)  #? [@width].pack("V")
        header[22,4] = lbo_dword(@height) #? [@height].pack("V")
        header[26] = 1.chr
        header[28] = 24.chr
        header[34,4] = lbo_dword(@size-54) #? [@size-54].pack("V")
        header[38,2] = 18.chr + 11.chr
        header[42,2] = 18.chr + 11.chr
        #
        begin
          #
          if @path.nil?
            file_path = file_name
          else
            file_path = File.join(@path,file_name)
          end
          #
          file = File.new(file_path,"wb")
          file.write(header)
          #
          padding = @width.divmod(4)[1]
          #
          for j in 0..(@height-1)
            for i in 0..(@width-1)
              b = @data[@height-1-j][i][2].chr
              g = @data[@height-1-j][i][1].chr
              r = @data[@height-1-j][i][0].chr
              file.write(b+g+r)
            end
            file.write(0.chr*padding)
          end
          #
        rescue => e
          #
          if @@debug
            puts()
            puts('SimpleRays - error writing file: '<<file_name)
            unless $VERBOSE.nil?
              puts("Error: #<#{e.class.name}: #{e.message}>")
              puts(e.backtrace) if $VERBOSE
            end
          end
          raise
          #
        ensure
          #
          file.close
          #
        end
      end #} write()

      #{ blur()
      #
      # !! Bad style: making assignments within boolean expressions !!
      #
      def blur(size)
        size.times do
          data = []
          for row in 0..@height - 1
            for col in 0..@width - 1
              average = [[0, 0, 0], 0]
              if (row > 0) and ((pixel = @data[row - 1][col]) != 0)
                average[0].offset!(pixel)
                average[1] += 1
              end
              if (row < @height - 1) and ((pixel = @data[row + 1][col]) != 0)
                average[0].offset!(pixel)
                average[1] += 1
              end
              if (col > 0) and ((pixel = @data[row][col - 1]) != 0)
                average[0].offset!(pixel)
                average[1] += 1
              end
              if (col < @width - 1) and ((pixel = @data[row][col + 1]) != 0)
                average[0].offset!(pixel)
                average[1] += 1
              end
              if (row > 0) and (col > 0) and ((pixel = @data[row - 1][col - 1]) != 0)
                average[0].offset!(pixel)
                average[1] += 1
              end
              if (row > 0) and (col < @width - 1) and ((pixel = @data[row - 1][col + 1]) != 0)
                average[0].offset!(pixel)
                average[1] += 1
              end
              if (row < @height - 1) and (col > 0) and ((pixel = @data[row + 1][col - 1]) != 0)
                average[0].offset!(pixel)
                average[1] += 1
              end
              if (row < @height - 1) and (col < @width - 1) and ((pixel = @data[row + 1][col + 1]) != 0)
                average[0].offset!(pixel)
                average[1] += 1
              end
              #
              if average[1] > 0
                t = Geom::Transformation.scaling( 1.0/average[1] )
                data << [ row, col, intized_array( average[0].transform(t) ) ]
              end
              #
            end
          end
          #
          for pixel in data
            # row = pixel[0]
            # col = pixel[1]
            # rgb = pixel[2]
            @data[ pixel[0] ][ pixel[1] ]= pixel[2]
          end
          #
        end # size.times
      end #} blur()

      private

      #{ intized_array( xyz_array )
      #
      #  Returns a new array whose members are the truncated
      #   values of the xyz_array argument.
      #--
      #  Equiv to Ruby's: array.map {|n| n.to_i }
      #++
      def intized_array( ary ) # called by blur()
        #
        return [ ary.x.to_i, ary.y.to_i, ary.z.to_i ]
        #
      end #} intized_array()


      #{ lbo_dword()
      #
      #  Outputs a 4-byte (32bit) DWORD String, in little-endian byte order.
      #
      # !! NEEDS OPTIMZING (outputs String)
      #    -> How about using Array.pack ?:  [n].pack("V")
      #
      def lbo_dword(n)
        n = n.to_i.to_s(16)
        n = '0' * (8 - n.length) + n # left pad with zeros
        n = n[6,2].to_i(16).chr+n[4,2].to_i(16).chr+n[2,2].to_i(16).chr+n[0,2].to_i(16).chr
      end #} lbo_dword()


      #{ from_byte() # <--<< NOT USED & mis-named: would take a 4 element array
      #
      # def from_byte(n)
        # n = (n[3].to_s(16)+n[2].to_s(16)+n[1].to_s(16)+n[0].to_s(16)).to_i(16)
      # end
      #} from_byte()


    end # class Bitmap


    class << self  # <--<< This module's anonymous singleton proxy class instance


      #{ debug( arg )
      #  debug=( arg )
      #
      #  If arg is a Numeric, any value greater than 0 is true. Otherwise
      #  sets the module's debug flag to the boolean evaluation of arg.
      #  If called with no argument, debug is set false.
      #
      def debug(arg=false)
        if arg.is_a?(Numeric)
          @@debug =( arg>0 ? true : false )
        else
          @@debug =( arg ? true : false )
        end
        @@debug_state =( @@debug ? MF_CHECKED : MF_UNCHECKED )
      end
      def debug=(arg) # alias
        debug(arg)
      end
      #} debug()


      #{ debug?
      #
      #  Returns boolean state of the debug flag.
      #
      def debug?()
        return @@debug
      end
      #} debug?()


      #{ exposure()
      #
      def exposure( xyz_array, vector_array, meth )
        model = Sketchup.active_model
        exp = 0  
        vector_array.each {|vector|
          if vector != 0
            exp += 1 if model.raytest([xyz_array, vector]).nil?
          end
        }
        if vector_array.length == 0
          rays = 1
        else
          rays = vector_array.length
        end
        exp = exp.to_f / rays
        #
        return rgb(exp,meth)
        #
      end #} exposure()


      #{ frontside_vector_array()
      #
      def frontside_vector_array(normal,vector_array)
        vector_array_front = []
        vector_array.each {|vector|
          if vector.dot(normal) > 0
            vector_array_front << vector
          else
            vector_array_front << 0 # is it smart to mix integer into an array of vector ??
          end
        }
        #
        return vector_array_front
        #
      end #} frontside_vector_array()


      #{ get_bounds_2d( array_of_xy_arrays )
      #
      #  Returns an array of form: [ [x_min,y_min], [x_max,y_max] ]
      #
      def get_bounds_2d( xy_arrays )
        # v1.0.0: replaced with built-in Array.min and Array.max
        return [
          [ xy_arrays.min {|a,b| a.x <=> b.x }.x, xy_arrays.min {|a,b| a.y <=> b.y }.y ],
          [ xy_arrays.max {|a,b| a.x <=> b.x }.x, xy_arrays.max {|a,b| a.y <=> b.y }.y ]
        ]
        #
      end #} get_bounds_2d()


      #{ get_point2d( xy_array, origin_array, vector_x, vector_y )
      #
      #  Returns a [x,y] array.
      #
      def get_point2d( point, origin, vx, vy )
        return [
          (origin.vector_to(point)).dot(vx),
          (origin.vector_to(point)).dot(vy)
        ]
      end #} get_point2d()


      #{ get_point3d( xy_point_array, xyz_origin_array, vector_x, vector_y )
      #
      #  If origin is an Array, returns a [x,y,z] array of Length class elements.
      #  If origin is a Geom::Point3d, returns a new Geom::Point3d object.
      #
      def get_point3d( point, origin, vx, vy )
        #
        return origin.offset( [
          vx.x * point.x + vy.x * point.y,
          vx.y * point.x + vy.y * point.y,
          vx.z * point.x + vy.z * point.y
        ])
        #
      end #} get_point3d()


      #{ hemisphere_vector_array()
      #
      def hemisphere_vector_array(normal, rays)
        vector_array = []
        vector_array << normal
        origin = Geom::Point3d.new( 0,0,0 )
        angle_vertical = 90.degrees / rays - 0.01
        angle_horizontal = 360.degrees / rays
        random_vector = [rand(100)+1, rand(100)+1, rand(100)+1]
        axis = normal * random_vector
        rotate_vertical = Geom::Transformation.rotation( origin, axis, angle_vertical )
        vector = normal.transform( rotate_vertical )
        #
        rays.times do
          rays.times do
            vector_array << vector
            rotate_horizontal = Geom::Transformation.rotation( origin, normal, angle_horizontal )
            vector = vector.transform( rotate_horizontal )
          end
          axis = normal * vector
          rotate_vertical = Geom::Transformation.rotation( origin, axis, angle_vertical )
          vector = vector.transform( rotate_vertical )
        end
        #
        return vector_array
        #
      end #} hemisphere_vector_array()


      #{ lightmap()
      #
      def lightmap(face, prec, progress, meth, vector_array, rays)
        normal = face.normal # ret: Geom::Vector3d
        # a line is: [ Geom::Point3d, Geom::Vector3d ]
        origin = face.edges[0].line[0].to_a # ret: Geom::Point3d.to_a()
        vx = face.edges[0].line[1]          # ret: Geom::Vector3d
        vy = vx * normal # ret: Geom::Vector3d perp to the plane of the vector operands
        #
        if meth == @@opt_text[:meth_0]
          # "Sun Exposure Time"
          vector_array = frontside_vector_array(normal,vector_array)
          #
        else
          # "Ambient Occlusion"
          vector_array = hemisphere_vector_array(normal, rays)
          #
        end
        #
        points2d = []
        for vert in face.vertices
          pt = get_point2d( vert.position, origin, vx, vy )
          points2d << [ round_f(pt.x,10), round_f(pt.y,10) ]
        end
        #
        # The method get_bounds_2d() returns an array
        #   of form: [ [x_min,y_min], [x_max,y_max] ]
        min2d,max2d = get_bounds_2d(points2d)
        #
        grid_x = (min2d.x-max2d.x).abs / prec
        grid_y = (min2d.y-max2d.y).abs / prec
        if grid_y > grid_x
          grid_y = grid_x
        else
          grid_x = grid_y
        end
        #
        min2d = [ min2d.x - grid_x    , min2d.y - grid_y     ]
        max2d = [ max2d.x + 2 * grid_x, max2d.y + 2 * grid_y ]
        #
        point2d  = [ min2d.x, min2d.y ] # need a new Array here!
        #
        position = [
          get_point3d( min2d, origin, vx, vy ),
          [0,1,0],
          get_point3d( [max2d.x - grid_x, max2d.y - grid_y], origin, vx, vy ),
          [1,0,0],
          get_point3d( [min2d.x , max2d.y - grid_y], origin, vx, vy ),
          [0,0,0]
        ]
        #
        lmap = []
        #
        percent = ((max2d.x - min2d.x) / grid_x * (max2d.y - min2d.y) / grid_x).abs * progress[1] / 100
        done = progress[0] * 100.to_f / progress[1]
        progress = 0 # progress was an Array, suddenly it is now an Integer -- ERROR-PRONE CODING STYLE !
        completed_old = (done + progress / percent).to_i
        #
        while point2d.y < max2d.y
          point2d.x= min2d.x
          row = []
          while point2d.x < max2d.x
            point = get_point3d( point2d, origin, vx, vy )
            if face.classify_point(point) == 1
              row << exposure( point, vector_array, meth )
            else
              row << 0
            end
            point2d.x= point2d.x + grid_x
            progress += 1
            completed = (done + progress / percent).to_i
            if completed != completed_old 
              Sketchup.set_status_text(completed.to_s + ' %')
              completed_old = completed
            end
          end
          lmap << row
          point2d.y= point2d.y + grid_y
        end
        #
        return [lmap, position]
        #
      end #} lightmap()

      
      #{ raytest()
      #
      def raytest(params)
        #
        return nil unless params # false if inputbox was cancelled
        #
        model = Sketchup.active_model
        seln = model.selection
        meth = params[0]
        prec = params[1].to_i
        rays = params[2].to_i
        mats = model.materials
        #
        if meth == @@opt_text[:meth_0] # "Sun Exposure Time"
          vector_array = sun_vector_array(rays)
        end
        #
        @@prevDir = Dir.getwd
        #
        model_path = model.path
        if model_path == ''
          ### v1.0.0
          #
          home = ENV['HOME'] || ENV['USERPROFILE']
          Dir.chdir(home)
          folder = '_temp_simple_rays'
          Dir.mkdir(folder) unless Kernel.test(?d,folder)
          Dir.chdir(folder)
          path = Dir.getwd
          folder = model.guid<<'-images'
          #
          # was saving images beneath the application binary path
          ##
        else
          folder = File.basename(model_path, ".skp") + '_images'
          path = File.dirname(model_path)
          Dir.chdir(path)
        end
        #
        Dir.mkdir(folder) unless Kernel.test(?d,folder) # was: if Dir[folder] == []
        Dir.chdir(folder)
        #
        faces = seln.grep(Sketchup::Face)
        faces_total = faces.length
        faces_done  = 0
        #
        for face in faces
          if @@debug
            @start = Time.now.to_f
            lmap = lightmap(face, prec, [faces_done, faces_total], meth, vector_array, rays)
            puts("lightmap #{faces_done+1} : #{Time.now.to_f - @start}")
          else
            lmap = lightmap(face, prec, [faces_done, faces_total], meth, vector_array, rays)
          end
          faces_done += 1
          bitmap = Bitmap.new(lmap[0])
          bitmap.blur(3)
          file = face.object_id.to_s + '.bmp'
          bitmap.path = File.join(path,folder) # v1.0.0
          bitmap.write(file)
          matl = mats.add( face.object_id.to_s )
          matl.texture = file
          face.position_material(matl, lmap[1], true)
        end # for
        #
        # materials.purge_unused
        #
      rescue => e
        #
        raise
        #
      ensure
        ### v1.0.0
        #
        unless @@debug
          #
          fpath = File.join(path,folder)
          # in case an observer changed the working dir:
          Dir.chdir(fpath) if Dir.getwd != fpath
          # cleanup files
          Dir.glob("*.{bmp}") {|file|
            File.delete(file) if Kernel.test(?f,file)
            #puts("Temp file: '#{file}'")
          }
          # cleanup the temp directory:
          if Dir[].empty? # avoid a SystemCallError
            Dir.chdir("..")
            Dir.rmdir(folder) if Kernel.test(?d,folder)
          end
          #
        end
        #
        Dir.chdir(@@prevDir) # restore the previous Dir
        #
        ###
      end #} raytest()


      #{ raytest_dialog()
      #
      def raytest_dialog()
        defaults = @@def_opts
        list = [@@raymeths,"",""]
        result = UI.inputbox(@@labels, defaults, list, @@menutext[:simple_rays])
        @@def_opts = result if result
        #
        return result
        #
      end #} raytest_dialog()


      #{ rgb()
      #
      def rgb( num, meth )
        if meth == @@opt_text[:meth_1]
          # "Ambient Occlusion"
          r = (255 * num).to_i
          g = (255 * num).to_i
          b = (255 * num).to_i
          #
        elsif meth == @@opt_text[:meth_0]
          # "Sun Exposure Time"
          if num < 0.3
            r = 0
            g = 0
            b = 255
          else
            r = (255 * (2 * num)).to_i
            if r < 0
              r = 0
            elsif r > 255
              r = 255
            end
            g = (255 * (2 * num - 1)).to_i
            if g < 0
              g = 0
            elsif g > 255
              g = 255
            end
            b = (255 * (1 - num)).to_i
            if b < 0
              b = 0
            elsif b > 255
              b = 255
            end
          end
          #
        end # if meth
        #
        return [r, g, b]
        #
      end #} rgb()


      #{ round_f( number, precision )
      #
      def round_f(n,x)
        (n * 10**x).round.to_f / 10**x
      end #} round_f()


      #{ rounded_array( xyz_array )
      #
      def rounded_array(ary,x)
        ary.map {|n| round_f(n,x) } 
      end #} rounded_array()


      #{ sun_vector_array()
      #
      def sun_vector_array(rays)
        model = Sketchup.active_model
        shadowinfo = model.shadow_info
        north = shadowinfo['NorthAngle'].degrees
        lat = shadowinfo['Latitude'].degrees
        lat = -lat
        angle = 15.degrees
        step = 150.degrees / rays
        vector_array = []
        (rays + 1).times do
          vector_array.push [Math.cos(angle) * Math.cos(north) + Math.sin(angle) * Math.sin(lat) * Math.sin(north), Math.sin(angle) * Math.sin(lat) * Math.cos(north) - Math.cos(angle) * Math.sin(north), Math.sin(angle) * Math.cos(lat)]
          angle += step
        end
        #
        return vector_array
        #
      end #} sun_vector_array()


    end # proxy class


    # ------------------------------------------------------------------------
    # UI features
    # ------------------------------------------------------------------------
    #{
    # Create the commands for the plugin on the context menu:
    # Only create commands if this file is not already loaded:
    #
    unless file_loaded?(File.join(BASEPATH, 'simple_rays.rb'))
    
      cmd = @@command[:simple_rays]= UI::Command.new("Simple Rays") {
        raytest(raytest_dialog)
      }
      cmd.menu_text= @@menutext[:simple_rays]
      cmd.set_validation_proc(&@@select_proc)

      # ----------------------------------------------------------------------
      # Create and add the SimpleRays submenu.
      #{----------------------------------------------------------------------
      @@submenu = UI.menu("Plugins").add_submenu( @@menutext[:submenu] )
      @@submenu.add_item @@command[:simple_rays]
      #
      @@submenu.add_separator
      # add the debug toggle:
      cmd = @@submenu.add_item(@@menutext[:debug_mode]){ debug(!@@debug) }
      @@submenu.set_validation_proc(cmd) { @@debug_state }
      #}

      # ----------------------------------------------------------------------
      # Create and add the SimpleRays context menu.
      #{----------------------------------------------------------------------
      UI.add_context_menu_handler {|menu|
        menu.add_separator
        menu.add_item @@command[:simple_rays]
      }

      # register the file as loaded:
      file_loaded(File.join(BASEPATH, 'simple_rays.rb'))

      cmd = nil # cleanup

    end #} if the file has not been loaded

  end # module SimpleRays

end # module Qpik